##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import numpy as np
import txaio

txaio.use_twisted()  # noqa

import pytest
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.xbrmm import Channel


def fill_channel(channel):
    channel.market_oid = uuid.uuid4()
    channel.member_oid = uuid.uuid4()
    channel.channel_oid = uuid.uuid4()
    channel.timestamp = np.datetime64(time_ns(), "ns")
    channel.open_at = random.randint(1, 2**256 - 1)
    channel.seq = random.randint(1, 2**32 - 1)
    channel.channel_type = random.randint(1, 2)
    channel.marketmaker = os.urandom(20)
    channel.actor = os.urandom(20)
    channel.delegate = os.urandom(20)
    channel.recipient = os.urandom(20)
    channel.amount = random.randint(1, 2**256 - 1)
    channel.timeout = random.randint(1, 2**32 - 1)
    channel.state = random.randint(1, 3)
    channel.closing_at = random.randint(1, 2**256 - 1)
    channel.closed_at = random.randint(1, 2**256 - 1)
    channel.close_mm_sig = os.urandom(65)
    channel.close_del_sig = os.urandom(65)
    channel.close_channel_seq = random.randint(1, 2**32 - 1)
    channel.close_is_final = random.choice([True, False])
    channel.close_balance = random.randint(1, 2**256 - 1)
    channel.closed_tx = os.urandom(32)


@pytest.fixture(scope="function")
def channel():
    _channel = Channel()
    fill_channel(_channel)
    return _channel


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_channel_roundtrip(channel, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = channel.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 720

    # create python object from bytes (flatbuffes)
    _channel = Channel.cast(data)

    assert _channel.market_oid == channel.market_oid
    assert _channel.member_oid == channel.member_oid
    assert _channel.channel_oid == channel.channel_oid
    assert _channel.timestamp == channel.timestamp
    assert _channel.open_at == channel.open_at
    assert _channel.seq == channel.seq
    assert _channel.channel_type == channel.channel_type
    assert _channel.marketmaker == channel.marketmaker
    assert _channel.actor == channel.actor
    assert _channel.delegate == channel.delegate
    assert _channel.recipient == channel.recipient
    assert _channel.amount == channel.amount
    assert _channel.timeout == channel.timeout
    assert _channel.state == channel.state
    assert _channel.closing_at == channel.closing_at
    assert _channel.closed_at == channel.closed_at
    assert _channel.close_mm_sig == channel.close_mm_sig
    assert _channel.close_del_sig == channel.close_del_sig
    assert _channel.close_channel_seq == channel.close_channel_seq
    assert _channel.close_is_final == channel.close_is_final
    assert _channel.close_balance == channel.close_balance
    assert _channel.closed_tx == channel.closed_tx


def test_channel_roundtrip_perf(channel, builder):
    obj = channel.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _channel = Channel.cast(data)
        if True:
            assert _channel.market_oid == channel.market_oid
            assert _channel.member_oid == channel.member_oid
            assert _channel.channel_oid == channel.channel_oid
            assert _channel.timestamp == channel.timestamp
            assert _channel.open_at == channel.open_at
            assert _channel.seq == channel.seq
            assert _channel.channel_type == channel.channel_type
            assert _channel.marketmaker == channel.marketmaker
            assert _channel.actor == channel.actor
            assert _channel.delegate == channel.delegate
            assert _channel.recipient == channel.recipient
            assert _channel.amount == channel.amount
            assert _channel.timeout == channel.timeout
            assert _channel.state == channel.state
            assert _channel.closing_at == channel.closing_at
            assert _channel.closed_at == channel.closed_at
            assert _channel.close_mm_sig == channel.close_mm_sig
            assert _channel.close_del_sig == channel.close_del_sig
            assert _channel.close_channel_seq == channel.close_channel_seq
            assert _channel.close_is_final == channel.close_is_final
            assert _channel.close_balance == channel.close_balance
            assert _channel.closed_tx == channel.closed_tx

            scratch["value"] += int(_channel.market_oid.bytes[0])
            scratch["value"] += int(_channel.market_oid.bytes[7])
            scratch["value"] += int(_channel.market_oid.bytes[15])

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import List, Optional
from uuid import UUID

import numpy as np

from cfxdb.mrealm.cluster import Cluster


class WebCluster(Cluster):
    """
    Web cluster database configuration object.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        name: Optional[str] = None,
        status: Optional[int] = None,
        owner_oid: Optional[UUID] = None,
        changed: Optional[np.datetime64] = None,
        tcp_version: Optional[int] = None,
        tcp_port: Optional[int] = None,
        tcp_shared: Optional[bool] = None,
        tcp_interface: Optional[str] = None,
        tcp_backlog: Optional[int] = None,
        tls_key: Optional[str] = None,
        tls_certificate: Optional[str] = None,
        tls_chain_certificates: Optional[List[str]] = None,
        tls_ca_certificates: Optional[List[str]] = None,
        tls_dhparam: Optional[str] = None,
        tls_ciphers: Optional[str] = None,
        http_client_timeout: Optional[int] = None,
        http_hsts: Optional[bool] = None,
        http_hsts_max_age: Optional[int] = None,
        http_access_log: Optional[bool] = None,
        http_display_tracebacks: Optional[bool] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of node

        :param label: Optional user label of node

        :param description: Optional user description of node

        :param tags: Optional list of user tags on node

        :param tcp_version: IP version, either 4 for 6

        :param tcp_port: IP listening port

        :param tcp_shared: enable TCP port sharing

        :param tcp_interface: listen on this interface

        :param tcp_backlog: TCP accept backlog queue size

        :param tls_key: TLS server private key to use

        :param tls_certificate: TLS server certificate to use

        :param tls_chain_certificates: TLS certificate chain

        :param tls_ca_certificates: CA certificates to use

        :param tls_dhparam: DH parameter file

        :param tls_ciphers: Ciphers list

        :param http_client_timeout: HTTP client inactivity timeout

        :param http_hsts: enable HTTP strict transport security (HSTS)

        :param http_hsts_max_age: HSTS maximum age to announce

        :param http_access_log: enable Web request access logging

        :param http_display_tracebacks: enable tracebacks when running into Web errors
        """
        Cluster.__init__(
            self,
            oid=oid,
            label=label,
            description=description,
            tags=tags,
            name=name,
            status=status,
            owner_oid=owner_oid,
            changed=changed,
            _unknown=_unknown,
        )
        self.tcp_version = tcp_version
        self.tcp_port = tcp_port
        self.tcp_shared = tcp_shared
        self.tcp_interface = tcp_interface
        self.tcp_backlog = tcp_backlog
        self.tls_key = tls_key
        self.tls_certificate = tls_certificate
        self.tls_chain_certificates = tls_chain_certificates
        self.tls_ca_certificates = tls_ca_certificates
        self.tls_dhparam = tls_dhparam
        self.tls_ciphers = tls_ciphers
        self.http_client_timeout = http_client_timeout
        self.http_hsts = http_hsts
        self.http_hsts_max_age = http_hsts_max_age
        self.http_access_log = http_access_log
        self.http_display_tracebacks = http_display_tracebacks

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not Cluster.__eq__(self, other):
            return False
        if other.tcp_version != self.tcp_version:
            return False
        if other.tcp_port != self.tcp_port:
            return False
        if other.tcp_shared != self.tcp_shared:
            return False
        if other.tcp_interface != self.tcp_interface:
            return False
        if other.tcp_backlog != self.tcp_backlog:
            return False
        if other.tls_key != self.tls_key:
            return False
        if other.tls_certificate != self.tls_certificate:
            return False
        if other.tls_chain_certificates != self.tls_chain_certificates:
            return False
        if other.tls_ca_certificates != self.tls_ca_certificates:
            return False
        if other.tls_dhparam != self.tls_dhparam:
            return False
        if other.tls_ciphers != self.tls_ciphers:
            return False
        if other.http_client_timeout != self.http_client_timeout:
            return False
        if other.http_hsts != self.http_hsts:
            return False
        if other.http_hsts_max_age != self.http_hsts_max_age:
            return False
        if other.http_access_log != self.http_access_log:
            return False
        if other.http_display_tracebacks != self.http_display_tracebacks:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = Cluster.marshal(self)

        obj.update(
            {
                "tcp_version": self.tcp_version,
                "tcp_port": self.tcp_port,
                "tcp_shared": self.tcp_shared,
                "tcp_interface": self.tcp_interface,
                "tcp_backlog": self.tcp_backlog,
                "tls_key": self.tls_key,
                "tls_certificate": self.tls_certificate,
                "tls_chain_certificates": self.tls_chain_certificates,
                "tls_ca_certificates": self.tls_ca_certificates,
                "tls_dhparam": self.tls_dhparam,
                "tls_ciphers": self.tls_ciphers,
                "http_client_timeout": self.http_client_timeout,
                "http_hsts": self.http_hsts,
                "http_hsts_max_age": self.http_hsts_max_age,
                "http_access_log": self.http_access_log,
                "http_display_tracebacks": self.http_display_tracebacks,
            }
        )

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        obj = Cluster.parse(data)
        data = obj._unknown or {}

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in [
                "tcp_version",
                "tcp_port",
                "tcp_shared",
                "tcp_interface",
                "tcp_backlog",
                "tls_key",
                "tls_certificate",
                "tls_chain_certificates",
                "tls_ca_certificates",
                "tls_dhparam",
                "tls_ciphers",
                "http_client_timeout",
                "http_hsts",
                "http_hsts_max_age",
                "http_access_log",
                "http_display_tracebacks",
            ]:
                _unknown[k] = data[k]

        tcp_version = data.get("tcp_version", None)
        assert tcp_version is None or (type(tcp_version) == int)

        tcp_port = data.get("tcp_port", None)
        assert tcp_port is None or (type(tcp_port) == int)

        tcp_shared = data.get("tcp_shared", None)
        assert tcp_shared is None or (type(tcp_shared) == bool)

        tcp_interface = data.get("tcp_interface", None)
        assert tcp_interface is None or (type(tcp_interface) == str)

        tcp_backlog = data.get("tcp_backlog", None)
        assert tcp_backlog is None or (type(tcp_backlog) == int)

        tls_key = data.get("tls_key", None)
        assert tls_key is None or (type(tls_key) == str)

        tls_certificate = data.get("tls_certificate", None)
        assert tls_certificate is None or (type(tls_certificate) == str)

        tls_chain_certificates = data.get("tls_chain_certificates", None)
        assert tls_chain_certificates is None or (type(tls_chain_certificates) == list)

        tls_ca_certificates = data.get("tls_ca_certificates", None)
        assert tls_ca_certificates is None or (type(tls_ca_certificates) == list)

        tls_dhparam = data.get("tls_dhparam", None)
        assert tls_dhparam is None or (type(tls_dhparam) == str)

        tls_ciphers = data.get("tls_ciphers", None)
        assert tls_ciphers is None or (type(tls_ciphers) == str)

        http_client_timeout = data.get("http_client_timeout", None)
        assert http_client_timeout is None or (type(http_client_timeout) == int)

        http_hsts = data.get("http_hsts", None)
        assert http_hsts is None or (type(http_hsts) == bool)

        http_hsts_max_age = data.get("http_hsts_max_age", None)
        assert http_hsts_max_age is None or (type(http_hsts_max_age) == int)

        http_access_log = data.get("http_access_log", None)
        assert http_access_log is None or (type(http_access_log) == bool)

        http_display_tracebacks = data.get("http_display_tracebacks", None)
        assert http_display_tracebacks is None or (type(http_display_tracebacks) == bool)

        obj = WebCluster(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            name=obj.name,
            status=obj.status,
            owner_oid=obj.owner_oid,
            changed=obj.changed,
            tcp_version=tcp_version,
            tcp_port=tcp_port,
            tcp_shared=tcp_shared,
            tcp_interface=tcp_interface,
            tcp_backlog=tcp_backlog,
            tls_key=tls_key,
            tls_certificate=tls_certificate,
            tls_chain_certificates=tls_chain_certificates,
            tls_ca_certificates=tls_ca_certificates,
            tls_dhparam=tls_dhparam,
            tls_ciphers=tls_ciphers,
            http_client_timeout=http_client_timeout,
            http_hsts=http_hsts,
            http_hsts_max_age=http_hsts_max_age,
            http_access_log=http_access_log,
            http_display_tracebacks=http_display_tracebacks,
            _unknown=_unknown,
        )

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import List, Optional
from uuid import UUID

import numpy as np

from cfxdb.common import ConfigurationElement
from cfxdb.gen.mrealm.ClusterStatus import ClusterStatus

STATUS_BY_CODE = {
    ClusterStatus.NONE: "NONE",
    ClusterStatus.STOPPED: "STOPPED",
    ClusterStatus.STARTING: "STARTING",
    ClusterStatus.RUNNING: "RUNNING",
    ClusterStatus.PAUSED: "PAUSED",
    ClusterStatus.STOPPING: "STOPPING",
    ClusterStatus.ERROR: "ERROR",
    ClusterStatus.DEGRADED: "DEGRADED",
}

STATUS_BY_NAME = {
    "NONE": ClusterStatus.NONE,
    "STOPPED": ClusterStatus.STOPPED,
    "STARTING": ClusterStatus.STARTING,
    "RUNNING": ClusterStatus.RUNNING,
    "PAUSED": ClusterStatus.PAUSED,
    "STOPPING": ClusterStatus.STOPPING,
    "ERROR": ClusterStatus.ERROR,
    "DEGRADED": ClusterStatus.DEGRADED,
}

STATUS_STOPPED = ClusterStatus.STOPPED
STATUS_STARTING = ClusterStatus.STARTING
STATUS_RUNNING = ClusterStatus.RUNNING
STATUS_PAUSED = ClusterStatus.PAUSED
STATUS_STOPPING = ClusterStatus.STOPPING
STATUS_ERROR = ClusterStatus.ERROR
STATUS_DEGRADED = ClusterStatus.DEGRADED


class Cluster(ConfigurationElement):
    """
    CFC Cluster database configuration object.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        name: Optional[str] = None,
        status: Optional[int] = None,
        owner_oid: Optional[UUID] = None,
        changed: Optional[np.datetime64] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of this cluster

        :param label: Optional user label of this cluster

        :param description: Optional user description of this cluster

        :param tags: Optional list of user tags on this cluster
        """
        ConfigurationElement.__init__(
            self, oid=oid, label=label, description=description, tags=tags, _unknown=_unknown
        )
        self.name = name
        self.status = status
        self.owner_oid = owner_oid
        self.changed = changed

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.name != self.name:
            return False
        if other.status != self.status:
            return False
        if other.owner_oid != self.owner_oid:
            return False
        if other.changed != self.changed:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ConfigurationElement.marshal(self)

        obj.update(
            {
                "name": self.name,
                "status": STATUS_BY_CODE[self.status] if self.status else None,
                "owner_oid": str(self.owner_oid) if self.owner_oid else None,
                "changed": int(self.changed) if self.changed else None,
            }
        )

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown or {}

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = dict()
        for k in data:
            if k not in ["name", "status", "owner_oid", "changed"]:
                _unknown[k] = data[k]

        name = data.get("name", None)
        assert name is None or (type(name) == str)

        status = data.get("status", None)
        assert status is None or (type(status) == str)
        status = STATUS_BY_NAME.get(status, None)

        owner_oid = data.get("owner_oid", None)
        assert owner_oid is None or (type(owner_oid) == str)
        if owner_oid:
            owner_oid = UUID(owner_oid)

        changed = data.get("changed", None)
        assert changed is None or (type(changed) == int)
        if changed:
            changed = np.datetime64(changed, "ns")

        obj = Cluster(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            name=name,
            status=status,
            owner_oid=owner_oid,
            changed=changed,
            _unknown=_unknown,
        )

        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: meta

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Generic **tags attached** to objects in other tables. Primary key of this table is ``(table_oid, object_oid, tag_oid)``.
class Tag(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Tag()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTag(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Tag
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Table of the object with the tag attached.
    # Tag
    def TableOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Tag
    def TableOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Tag
    def TableOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tag
    def TableOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Object (within the table) with the tag attached.
    # Tag
    def ObjectOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Tag
    def ObjectOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Tag
    def ObjectOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tag
    def ObjectOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Tag definition attached to the object.
    # Tag
    def TagOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Tag
    def TagOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Tag
    def TagOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tag
    def TagOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Timestamp when the tag was attached to the object referenced.
    # Tag
    def Attached(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0


def TagStart(builder):
    builder.StartObject(4)


def Start(builder):
    TagStart(builder)


def TagAddTableOid(builder, tableOid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(tableOid), 0)


def AddTableOid(builder, tableOid):
    TagAddTableOid(builder, tableOid)


def TagStartTableOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTableOidVector(builder, numElems: int) -> int:
    return TagStartTableOidVector(builder, numElems)


def TagAddObjectOid(builder, objectOid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(objectOid), 0)


def AddObjectOid(builder, objectOid):
    TagAddObjectOid(builder, objectOid)


def TagStartObjectOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartObjectOidVector(builder, numElems: int) -> int:
    return TagStartObjectOidVector(builder, numElems)


def TagAddTagOid(builder, tagOid):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(tagOid), 0)


def AddTagOid(builder, tagOid):
    TagAddTagOid(builder, tagOid)


def TagStartTagOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTagOidVector(builder, numElems: int) -> int:
    return TagStartTagOidVector(builder, numElems)


def TagAddAttached(builder, attached):
    builder.PrependUint64Slot(3, attached, 0)


def AddAttached(builder, attached):
    TagAddAttached(builder, attached)


def TagEnd(builder):
    return builder.EndObject()


def End(builder):
    return TagEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Heartbeat log record of managed nodes. Primary key: (timestamp, node_id).
class MNodeLog(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MNodeLog()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMNodeLog(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # MNodeLog
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Primary key: Unix time in ns when this log record was received (from CFC node clock).
    # MNodeLog
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Primary key: CF node ID.
    # MNodeLog
    def NodeId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MNodeLog
    def NodeIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MNodeLog
    def NodeIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MNodeLog
    def NodeIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Heartbeat period in seconds encompassed by this record.
    # MNodeLog
    def Period(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Management realm ID.
    # MNodeLog
    def MrealmId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MNodeLog
    def MrealmIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MNodeLog
    def MrealmIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MNodeLog
    def MrealmIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Current state of CF node.
    # MNodeLog
    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # WAMP session ID of the CF node uplink management session to this CFC instance.
    # MNodeLog
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Unix time in ns. This timestamp is from the original received event payload (from CF node clock).
    # MNodeLog
    def Sent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Sequence number as sent in the log record by the CF node (started at 0 for CF start and incremented by one on each heartbeat).
    # MNodeLog
    def Seq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Number of controllers running in the CF node (this is always 1).
    # MNodeLog
    def Controllers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Number of router workers currently running in the CF node.
    # MNodeLog
    def Routers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Number of container workers currently running in the CF node.
    # MNodeLog
    def Containers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Number of guest workers currently running in the CF node.
    # MNodeLog
    def Guests(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Number of proxy workers currently running in the CF node.
    # MNodeLog
    def Proxies(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Number of XBR market maker workers currently running in the CF node.
    # MNodeLog
    def Marketmakers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Number of hostmonitor workers currently running in the CF node (this is usually either 0 or 1).
    # MNodeLog
    def Hostmonitors(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # CF node system statistics.
    # MNodeLog
    def CpuCtxSwitches(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def CpuFreq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuGuest(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuGuestNice(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuIdle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuInterrupts(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def CpuIowait(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuIrq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(48))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuNice(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(50))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuSoftInterrupts(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(52))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def CpuSoftirq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(54))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuSteal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(56))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuSystem(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(58))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def CpuUser(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(60))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def DiskBusyTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(62))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def DiskReadBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(64))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def DiskReadCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(66))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def DiskReadMergedCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(68))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def DiskReadTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(70))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def DiskWriteBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(72))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def DiskWriteCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(74))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def DiskWriteMergedCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(76))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def DiskWriteTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(78))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryActive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(80))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryAvailable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(82))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryBuffers(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(84))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryCached(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(86))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryFree(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(88))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryInactive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(90))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryPercent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(92))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # MNodeLog
    def MemoryShared(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(94))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemorySlab(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(96))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryTotal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(98))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def MemoryUsed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(100))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkBytesRecv(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(102))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkBytesSent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(104))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkConnectionAfInet(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(106))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkConnectionAfInet6(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(108))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkConnectionAfUnix(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(110))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkDropin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(112))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkDropout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(114))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkErrin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(116))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkErrout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(118))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkPacketsRecv(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(120))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # MNodeLog
    def NetworkPacketsSent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(122))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0


def MNodeLogStart(builder):
    builder.StartObject(60)


def Start(builder):
    MNodeLogStart(builder)


def MNodeLogAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def AddTimestamp(builder, timestamp):
    MNodeLogAddTimestamp(builder, timestamp)


def MNodeLogAddNodeId(builder, nodeId):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(nodeId), 0)


def AddNodeId(builder, nodeId):
    MNodeLogAddNodeId(builder, nodeId)


def MNodeLogStartNodeIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartNodeIdVector(builder, numElems: int) -> int:
    return MNodeLogStartNodeIdVector(builder, numElems)


def MNodeLogAddPeriod(builder, period):
    builder.PrependUint32Slot(2, period, 0)


def AddPeriod(builder, period):
    MNodeLogAddPeriod(builder, period)


def MNodeLogAddMrealmId(builder, mrealmId):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(mrealmId), 0)


def AddMrealmId(builder, mrealmId):
    MNodeLogAddMrealmId(builder, mrealmId)


def MNodeLogStartMrealmIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMrealmIdVector(builder, numElems: int) -> int:
    return MNodeLogStartMrealmIdVector(builder, numElems)


def MNodeLogAddState(builder, state):
    builder.PrependUint8Slot(4, state, 0)


def AddState(builder, state):
    MNodeLogAddState(builder, state)


def MNodeLogAddSession(builder, session):
    builder.PrependUint64Slot(5, session, 0)


def AddSession(builder, session):
    MNodeLogAddSession(builder, session)


def MNodeLogAddSent(builder, sent):
    builder.PrependUint64Slot(6, sent, 0)


def AddSent(builder, sent):
    MNodeLogAddSent(builder, sent)


def MNodeLogAddSeq(builder, seq):
    builder.PrependUint64Slot(7, seq, 0)


def AddSeq(builder, seq):
    MNodeLogAddSeq(builder, seq)


def MNodeLogAddControllers(builder, controllers):
    builder.PrependUint16Slot(8, controllers, 0)


def AddControllers(builder, controllers):
    MNodeLogAddControllers(builder, controllers)


def MNodeLogAddRouters(builder, routers):
    builder.PrependUint16Slot(9, routers, 0)


def AddRouters(builder, routers):
    MNodeLogAddRouters(builder, routers)


def MNodeLogAddContainers(builder, containers):
    builder.PrependUint16Slot(10, containers, 0)


def AddContainers(builder, containers):
    MNodeLogAddContainers(builder, containers)


def MNodeLogAddGuests(builder, guests):
    builder.PrependUint16Slot(11, guests, 0)


def AddGuests(builder, guests):
    MNodeLogAddGuests(builder, guests)


def MNodeLogAddProxies(builder, proxies):
    builder.PrependUint16Slot(12, proxies, 0)


def AddProxies(builder, proxies):
    MNodeLogAddProxies(builder, proxies)


def MNodeLogAddMarketmakers(builder, marketmakers):
    builder.PrependUint16Slot(13, marketmakers, 0)


def AddMarketmakers(builder, marketmakers):
    MNodeLogAddMarketmakers(builder, marketmakers)


def MNodeLogAddHostmonitors(builder, hostmonitors):
    builder.PrependUint16Slot(14, hostmonitors, 0)


def AddHostmonitors(builder, hostmonitors):
    MNodeLogAddHostmonitors(builder, hostmonitors)


def MNodeLogAddCpuCtxSwitches(builder, cpuCtxSwitches):
    builder.PrependUint64Slot(15, cpuCtxSwitches, 0)


def AddCpuCtxSwitches(builder, cpuCtxSwitches):
    MNodeLogAddCpuCtxSwitches(builder, cpuCtxSwitches)


def MNodeLogAddCpuFreq(builder, cpuFreq):
    builder.PrependFloat32Slot(16, cpuFreq, 0.0)


def AddCpuFreq(builder, cpuFreq):
    MNodeLogAddCpuFreq(builder, cpuFreq)


def MNodeLogAddCpuGuest(builder, cpuGuest):
    builder.PrependFloat32Slot(17, cpuGuest, 0.0)


def AddCpuGuest(builder, cpuGuest):
    MNodeLogAddCpuGuest(builder, cpuGuest)


def MNodeLogAddCpuGuestNice(builder, cpuGuestNice):
    builder.PrependFloat32Slot(18, cpuGuestNice, 0.0)


def AddCpuGuestNice(builder, cpuGuestNice):
    MNodeLogAddCpuGuestNice(builder, cpuGuestNice)


def MNodeLogAddCpuIdle(builder, cpuIdle):
    builder.PrependFloat32Slot(19, cpuIdle, 0.0)


def AddCpuIdle(builder, cpuIdle):
    MNodeLogAddCpuIdle(builder, cpuIdle)


def MNodeLogAddCpuInterrupts(builder, cpuInterrupts):
    builder.PrependUint64Slot(20, cpuInterrupts, 0)


def AddCpuInterrupts(builder, cpuInterrupts):
    MNodeLogAddCpuInterrupts(builder, cpuInterrupts)


def MNodeLogAddCpuIowait(builder, cpuIowait):
    builder.PrependFloat32Slot(21, cpuIowait, 0.0)


def AddCpuIowait(builder, cpuIowait):
    MNodeLogAddCpuIowait(builder, cpuIowait)


def MNodeLogAddCpuIrq(builder, cpuIrq):
    builder.PrependFloat32Slot(22, cpuIrq, 0.0)


def AddCpuIrq(builder, cpuIrq):
    MNodeLogAddCpuIrq(builder, cpuIrq)


def MNodeLogAddCpuNice(builder, cpuNice):
    builder.PrependFloat32Slot(23, cpuNice, 0.0)


def AddCpuNice(builder, cpuNice):
    MNodeLogAddCpuNice(builder, cpuNice)


def MNodeLogAddCpuSoftInterrupts(builder, cpuSoftInterrupts):
    builder.PrependUint64Slot(24, cpuSoftInterrupts, 0)


def AddCpuSoftInterrupts(builder, cpuSoftInterrupts):
    MNodeLogAddCpuSoftInterrupts(builder, cpuSoftInterrupts)


def MNodeLogAddCpuSoftirq(builder, cpuSoftirq):
    builder.PrependFloat32Slot(25, cpuSoftirq, 0.0)


def AddCpuSoftirq(builder, cpuSoftirq):
    MNodeLogAddCpuSoftirq(builder, cpuSoftirq)


def MNodeLogAddCpuSteal(builder, cpuSteal):
    builder.PrependFloat32Slot(26, cpuSteal, 0.0)


def AddCpuSteal(builder, cpuSteal):
    MNodeLogAddCpuSteal(builder, cpuSteal)


def MNodeLogAddCpuSystem(builder, cpuSystem):
    builder.PrependFloat32Slot(27, cpuSystem, 0.0)


def AddCpuSystem(builder, cpuSystem):
    MNodeLogAddCpuSystem(builder, cpuSystem)


def MNodeLogAddCpuUser(builder, cpuUser):
    builder.PrependFloat32Slot(28, cpuUser, 0.0)


def AddCpuUser(builder, cpuUser):
    MNodeLogAddCpuUser(builder, cpuUser)


def MNodeLogAddDiskBusyTime(builder, diskBusyTime):
    builder.PrependUint64Slot(29, diskBusyTime, 0)


def AddDiskBusyTime(builder, diskBusyTime):
    MNodeLogAddDiskBusyTime(builder, diskBusyTime)


def MNodeLogAddDiskReadBytes(builder, diskReadBytes):
    builder.PrependUint64Slot(30, diskReadBytes, 0)


def AddDiskReadBytes(builder, diskReadBytes):
    MNodeLogAddDiskReadBytes(builder, diskReadBytes)


def MNodeLogAddDiskReadCount(builder, diskReadCount):
    builder.PrependUint64Slot(31, diskReadCount, 0)


def AddDiskReadCount(builder, diskReadCount):
    MNodeLogAddDiskReadCount(builder, diskReadCount)


def MNodeLogAddDiskReadMergedCount(builder, diskReadMergedCount):
    builder.PrependUint64Slot(32, diskReadMergedCount, 0)


def AddDiskReadMergedCount(builder, diskReadMergedCount):
    MNodeLogAddDiskReadMergedCount(builder, diskReadMergedCount)


def MNodeLogAddDiskReadTime(builder, diskReadTime):
    builder.PrependUint64Slot(33, diskReadTime, 0)


def AddDiskReadTime(builder, diskReadTime):
    MNodeLogAddDiskReadTime(builder, diskReadTime)


def MNodeLogAddDiskWriteBytes(builder, diskWriteBytes):
    builder.PrependUint64Slot(34, diskWriteBytes, 0)


def AddDiskWriteBytes(builder, diskWriteBytes):
    MNodeLogAddDiskWriteBytes(builder, diskWriteBytes)


def MNodeLogAddDiskWriteCount(builder, diskWriteCount):
    builder.PrependUint64Slot(35, diskWriteCount, 0)


def AddDiskWriteCount(builder, diskWriteCount):
    MNodeLogAddDiskWriteCount(builder, diskWriteCount)


def MNodeLogAddDiskWriteMergedCount(builder, diskWriteMergedCount):
    builder.PrependUint64Slot(36, diskWriteMergedCount, 0)


def AddDiskWriteMergedCount(builder, diskWriteMergedCount):
    MNodeLogAddDiskWriteMergedCount(builder, diskWriteMergedCount)


def MNodeLogAddDiskWriteTime(builder, diskWriteTime):
    builder.PrependUint64Slot(37, diskWriteTime, 0)


def AddDiskWriteTime(builder, diskWriteTime):
    MNodeLogAddDiskWriteTime(builder, diskWriteTime)


def MNodeLogAddMemoryActive(builder, memoryActive):
    builder.PrependUint64Slot(38, memoryActive, 0)


def AddMemoryActive(builder, memoryActive):
    MNodeLogAddMemoryActive(builder, memoryActive)


def MNodeLogAddMemoryAvailable(builder, memoryAvailable):
    builder.PrependUint64Slot(39, memoryAvailable, 0)


def AddMemoryAvailable(builder, memoryAvailable):
    MNodeLogAddMemoryAvailable(builder, memoryAvailable)


def MNodeLogAddMemoryBuffers(builder, memoryBuffers):
    builder.PrependUint64Slot(40, memoryBuffers, 0)


def AddMemoryBuffers(builder, memoryBuffers):
    MNodeLogAddMemoryBuffers(builder, memoryBuffers)


def MNodeLogAddMemoryCached(builder, memoryCached):
    builder.PrependUint64Slot(41, memoryCached, 0)


def AddMemoryCached(builder, memoryCached):
    MNodeLogAddMemoryCached(builder, memoryCached)


def MNodeLogAddMemoryFree(builder, memoryFree):
    builder.PrependUint64Slot(42, memoryFree, 0)


def AddMemoryFree(builder, memoryFree):
    MNodeLogAddMemoryFree(builder, memoryFree)


def MNodeLogAddMemoryInactive(builder, memoryInactive):
    builder.PrependUint64Slot(43, memoryInactive, 0)


def AddMemoryInactive(builder, memoryInactive):
    MNodeLogAddMemoryInactive(builder, memoryInactive)


def MNodeLogAddMemoryPercent(builder, memoryPercent):
    builder.PrependFloat32Slot(44, memoryPercent, 0.0)


def AddMemoryPercent(builder, memoryPercent):
    MNodeLogAddMemoryPercent(builder, memoryPercent)


def MNodeLogAddMemoryShared(builder, memoryShared):
    builder.PrependUint64Slot(45, memoryShared, 0)


def AddMemoryShared(builder, memoryShared):
    MNodeLogAddMemoryShared(builder, memoryShared)


def MNodeLogAddMemorySlab(builder, memorySlab):
    builder.PrependUint64Slot(46, memorySlab, 0)


def AddMemorySlab(builder, memorySlab):
    MNodeLogAddMemorySlab(builder, memorySlab)


def MNodeLogAddMemoryTotal(builder, memoryTotal):
    builder.PrependUint64Slot(47, memoryTotal, 0)


def AddMemoryTotal(builder, memoryTotal):
    MNodeLogAddMemoryTotal(builder, memoryTotal)


def MNodeLogAddMemoryUsed(builder, memoryUsed):
    builder.PrependUint64Slot(48, memoryUsed, 0)


def AddMemoryUsed(builder, memoryUsed):
    MNodeLogAddMemoryUsed(builder, memoryUsed)


def MNodeLogAddNetworkBytesRecv(builder, networkBytesRecv):
    builder.PrependUint64Slot(49, networkBytesRecv, 0)


def AddNetworkBytesRecv(builder, networkBytesRecv):
    MNodeLogAddNetworkBytesRecv(builder, networkBytesRecv)


def MNodeLogAddNetworkBytesSent(builder, networkBytesSent):
    builder.PrependUint64Slot(50, networkBytesSent, 0)


def AddNetworkBytesSent(builder, networkBytesSent):
    MNodeLogAddNetworkBytesSent(builder, networkBytesSent)


def MNodeLogAddNetworkConnectionAfInet(builder, networkConnectionAfInet):
    builder.PrependUint32Slot(51, networkConnectionAfInet, 0)


def AddNetworkConnectionAfInet(builder, networkConnectionAfInet):
    MNodeLogAddNetworkConnectionAfInet(builder, networkConnectionAfInet)


def MNodeLogAddNetworkConnectionAfInet6(builder, networkConnectionAfInet6):
    builder.PrependUint32Slot(52, networkConnectionAfInet6, 0)


def AddNetworkConnectionAfInet6(builder, networkConnectionAfInet6):
    MNodeLogAddNetworkConnectionAfInet6(builder, networkConnectionAfInet6)


def MNodeLogAddNetworkConnectionAfUnix(builder, networkConnectionAfUnix):
    builder.PrependUint32Slot(53, networkConnectionAfUnix, 0)


def AddNetworkConnectionAfUnix(builder, networkConnectionAfUnix):
    MNodeLogAddNetworkConnectionAfUnix(builder, networkConnectionAfUnix)


def MNodeLogAddNetworkDropin(builder, networkDropin):
    builder.PrependUint32Slot(54, networkDropin, 0)


def AddNetworkDropin(builder, networkDropin):
    MNodeLogAddNetworkDropin(builder, networkDropin)


def MNodeLogAddNetworkDropout(builder, networkDropout):
    builder.PrependUint32Slot(55, networkDropout, 0)


def AddNetworkDropout(builder, networkDropout):
    MNodeLogAddNetworkDropout(builder, networkDropout)


def MNodeLogAddNetworkErrin(builder, networkErrin):
    builder.PrependUint32Slot(56, networkErrin, 0)


def AddNetworkErrin(builder, networkErrin):
    MNodeLogAddNetworkErrin(builder, networkErrin)


def MNodeLogAddNetworkErrout(builder, networkErrout):
    builder.PrependUint32Slot(57, networkErrout, 0)


def AddNetworkErrout(builder, networkErrout):
    MNodeLogAddNetworkErrout(builder, networkErrout)


def MNodeLogAddNetworkPacketsRecv(builder, networkPacketsRecv):
    builder.PrependUint64Slot(58, networkPacketsRecv, 0)


def AddNetworkPacketsRecv(builder, networkPacketsRecv):
    MNodeLogAddNetworkPacketsRecv(builder, networkPacketsRecv)


def MNodeLogAddNetworkPacketsSent(builder, networkPacketsSent):
    builder.PrependUint64Slot(59, networkPacketsSent, 0)


def AddNetworkPacketsSent(builder, networkPacketsSent):
    MNodeLogAddNetworkPacketsSent(builder, networkPacketsSent)


def MNodeLogEnd(builder):
    return builder.EndObject()


def End(builder):
    return MNodeLogEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: realmstore

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# This table stores WAMP sessions and serves as an anchor for all usage related data.
class Session(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Session()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSession(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Session
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OID of the application realm this session is/was joined on.
    # Session
    def ArealmOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Session
    def ArealmOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Session
    def ArealmOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Session
    def ArealmOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Unlimited time, globally unique, long-term OID of this session. The pair of WAMP session ID and join time ``(session, joined_at)`` bidirectionally maps to session ``oid``.
    # Session
    def Oid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Session
    def OidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Session
    def OidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Session
    def OidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # The WAMP session_id of the session.
    # Session
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Timestamp when the session was joined by the router. Epoch time in ns.
    # Session
    def JoinedAt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Timestamp when the session left the router. Epoch time in ns.
    # Session
    def LeftAt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # OID of the node of the router worker hosting this session.
    # Session
    def NodeOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Session
    def NodeOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Session
    def NodeOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Session
    def NodeOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Name (management realm WAMP authid) of the node of the router worker hosting this session.
    # Session
    def NodeAuthid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Local worker name of the router worker hosting this session.
    # Session
    def WorkerName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Local worker PID of the router worker hosting this session.
    # Session
    def WorkerPid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Session transport information, the incoming frontend client connection in proxy-router setups. This is also returned as part of authextra to the client.
    # Session
    def Transport(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Session
    def TransportAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Session
    def TransportLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Session
    def TransportIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

    # From proxy (in proxy-router cluster setups): OID of the node of the proxy worker hosting this session.
    # Session
    def ProxyNodeOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Session
    def ProxyNodeOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Session
    def ProxyNodeOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Session
    def ProxyNodeOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # From proxy (in proxy-router cluster setups): Name (management realm WAMP authid) of the node of the proxy worker hosting this session.
    # Session
    def ProxyNodeAuthid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # From proxy (in proxy-router cluster setups): Local worker name of the proxy worker hosting this session.
    # Session
    def ProxyWorkerName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # From proxy (in proxy-router cluster setups): Local worker PID of the proxy worker hosting this session.
    # Session
    def ProxyWorkerPid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # From proxy (in proxy-router cluster setups): Session transport information, the transport from the proxy to the backend router.
    # Session
    def ProxyTransport(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Session
    def ProxyTransportAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Session
    def ProxyTransportLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Session
    def ProxyTransportIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        return o == 0

    # The WAMP realm (name) the session is/was joined on.
    # Session
    def Realm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authid the session was authenticated under.
    # Session
    def Authid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authrole the session was authenticated under.
    # Session
    def Authrole(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authmethod uses to authenticate the session.
    # Session
    def Authmethod(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authprovider that was handling the session authentication.
    # Session
    def Authprovider(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authextra as provided to the authenticated session.
    # Session
    def Authextra(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Session
    def AuthextraAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Session
    def AuthextraLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Session
    def AuthextraIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        return o == 0


def SessionStart(builder):
    builder.StartObject(21)


def Start(builder):
    SessionStart(builder)


def SessionAddArealmOid(builder, arealmOid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(arealmOid), 0)


def AddArealmOid(builder, arealmOid):
    SessionAddArealmOid(builder, arealmOid)


def SessionStartArealmOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartArealmOidVector(builder, numElems: int) -> int:
    return SessionStartArealmOidVector(builder, numElems)


def SessionAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    SessionAddOid(builder, oid)


def SessionStartOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOidVector(builder, numElems: int) -> int:
    return SessionStartOidVector(builder, numElems)


def SessionAddSession(builder, session):
    builder.PrependUint64Slot(2, session, 0)


def AddSession(builder, session):
    SessionAddSession(builder, session)


def SessionAddJoinedAt(builder, joinedAt):
    builder.PrependUint64Slot(3, joinedAt, 0)


def AddJoinedAt(builder, joinedAt):
    SessionAddJoinedAt(builder, joinedAt)


def SessionAddLeftAt(builder, leftAt):
    builder.PrependUint64Slot(4, leftAt, 0)


def AddLeftAt(builder, leftAt):
    SessionAddLeftAt(builder, leftAt)


def SessionAddNodeOid(builder, nodeOid):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(nodeOid), 0)


def AddNodeOid(builder, nodeOid):
    SessionAddNodeOid(builder, nodeOid)


def SessionStartNodeOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartNodeOidVector(builder, numElems: int) -> int:
    return SessionStartNodeOidVector(builder, numElems)


def SessionAddNodeAuthid(builder, nodeAuthid):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(nodeAuthid), 0)


def AddNodeAuthid(builder, nodeAuthid):
    SessionAddNodeAuthid(builder, nodeAuthid)


def SessionAddWorkerName(builder, workerName):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(workerName), 0)


def AddWorkerName(builder, workerName):
    SessionAddWorkerName(builder, workerName)


def SessionAddWorkerPid(builder, workerPid):
    builder.PrependInt32Slot(8, workerPid, 0)


def AddWorkerPid(builder, workerPid):
    SessionAddWorkerPid(builder, workerPid)


def SessionAddTransport(builder, transport):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(transport), 0)


def AddTransport(builder, transport):
    SessionAddTransport(builder, transport)


def SessionStartTransportVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTransportVector(builder, numElems: int) -> int:
    return SessionStartTransportVector(builder, numElems)


def SessionAddProxyNodeOid(builder, proxyNodeOid):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(proxyNodeOid), 0)


def AddProxyNodeOid(builder, proxyNodeOid):
    SessionAddProxyNodeOid(builder, proxyNodeOid)


def SessionStartProxyNodeOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartProxyNodeOidVector(builder, numElems: int) -> int:
    return SessionStartProxyNodeOidVector(builder, numElems)


def SessionAddProxyNodeAuthid(builder, proxyNodeAuthid):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(proxyNodeAuthid), 0)


def AddProxyNodeAuthid(builder, proxyNodeAuthid):
    SessionAddProxyNodeAuthid(builder, proxyNodeAuthid)


def SessionAddProxyWorkerName(builder, proxyWorkerName):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(proxyWorkerName), 0)


def AddProxyWorkerName(builder, proxyWorkerName):
    SessionAddProxyWorkerName(builder, proxyWorkerName)


def SessionAddProxyWorkerPid(builder, proxyWorkerPid):
    builder.PrependInt32Slot(13, proxyWorkerPid, 0)


def AddProxyWorkerPid(builder, proxyWorkerPid):
    SessionAddProxyWorkerPid(builder, proxyWorkerPid)


def SessionAddProxyTransport(builder, proxyTransport):
    builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(proxyTransport), 0)


def AddProxyTransport(builder, proxyTransport):
    SessionAddProxyTransport(builder, proxyTransport)


def SessionStartProxyTransportVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartProxyTransportVector(builder, numElems: int) -> int:
    return SessionStartProxyTransportVector(builder, numElems)


def SessionAddRealm(builder, realm):
    builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(realm), 0)


def AddRealm(builder, realm):
    SessionAddRealm(builder, realm)


def SessionAddAuthid(builder, authid):
    builder.PrependUOffsetTRelativeSlot(16, flatbuffers.number_types.UOffsetTFlags.py_type(authid), 0)


def AddAuthid(builder, authid):
    SessionAddAuthid(builder, authid)


def SessionAddAuthrole(builder, authrole):
    builder.PrependUOffsetTRelativeSlot(17, flatbuffers.number_types.UOffsetTFlags.py_type(authrole), 0)


def AddAuthrole(builder, authrole):
    SessionAddAuthrole(builder, authrole)


def SessionAddAuthmethod(builder, authmethod):
    builder.PrependUOffsetTRelativeSlot(18, flatbuffers.number_types.UOffsetTFlags.py_type(authmethod), 0)


def AddAuthmethod(builder, authmethod):
    SessionAddAuthmethod(builder, authmethod)


def SessionAddAuthprovider(builder, authprovider):
    builder.PrependUOffsetTRelativeSlot(19, flatbuffers.number_types.UOffsetTFlags.py_type(authprovider), 0)


def AddAuthprovider(builder, authprovider):
    SessionAddAuthprovider(builder, authprovider)


def SessionAddAuthextra(builder, authextra):
    builder.PrependUOffsetTRelativeSlot(20, flatbuffers.number_types.UOffsetTFlags.py_type(authextra), 0)


def AddAuthextra(builder, authextra):
    SessionAddAuthextra(builder, authextra)


def SessionStartAuthextraVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAuthextraVector(builder, numElems: int) -> int:
    return SessionStartAuthextraVector(builder, numElems)


def SessionEnd(builder):
    return builder.EndObject()


def End(builder):
    return SessionEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import List, Optional
from uuid import UUID

import numpy as np

from cfxdb.common import ConfigurationElement


class Permission(ConfigurationElement):
    """
    Role permission database object.
    """

    MATCH_TYPE_NONE: int = 0
    MATCH_TYPE_EXACT: int = 1
    MATCH_TYPE_PREFIX: int = 2
    MATCH_TYPE_WILDCARD: int = 3
    MATCH_TYPES_TOSTR = {
        0: None,
        1: "exact",
        2: "prefix",
        3: "wildcard",
    }
    MATCH_TYPES_FROMSTR = {val: key for key, val in MATCH_TYPES_TOSTR.items()}

    URI_CHECK_LEVEL_NONE: int = 0
    URI_CHECK_LEVEL_STRICT: int = 1
    URI_CHECK_LEVEL_LOOSE: int = 2
    URI_CHECK_LEVELS_TOSTR = {0: None, 1: "strict", 2: "loose"}
    URI_CHECK_LEVELS_FROMSTR = {val: key for key, val in URI_CHECK_LEVELS_TOSTR.items()}

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        role_oid: Optional[UUID] = None,
        uri: Optional[str] = None,
        uri_check_level: Optional[int] = None,
        match: Optional[int] = None,
        allow_call: Optional[bool] = None,
        allow_register: Optional[bool] = None,
        allow_publish: Optional[bool] = None,
        allow_subscribe: Optional[bool] = None,
        disclose_caller: Optional[bool] = None,
        disclose_publisher: Optional[bool] = None,
        cache: Optional[bool] = None,
        created: Optional[np.datetime64] = None,
        owner: Optional[UUID] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of this permission object

        :param label: Optional user label of permission

        :param description: Optional user description of permission

        :param tags: Optional list of user tags on permission

        :param role_oid: Object ID of role this permission applies to.

        :param uri: URI matched for permission.

        :param created: Timestamp when the permission was created

        :param owner: Owning user (object ID)
        """
        ConfigurationElement.__init__(self, oid=oid, label=label, description=description, tags=tags)
        self.role_oid = role_oid
        self.uri = uri
        self.uri_check_level = uri_check_level
        self.match = match
        self.allow_call = allow_call
        self.allow_register = allow_register
        self.allow_publish = allow_publish
        self.allow_subscribe = allow_subscribe
        self.disclose_caller = disclose_caller
        self.disclose_publisher = disclose_publisher
        self.cache = cache
        self.created = created
        self.owner = owner

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.role_oid != self.role_oid:
            return False
        if other.uri != self.uri:
            return False
        if other.uri_check_level != self.uri_check_level:
            return False
        if other.match != self.match:
            return False
        if other.allow_call != self.allow_call:
            return False
        if other.allow_register != self.allow_register:
            return False
        if other.allow_publish != self.allow_publish:
            return False
        if other.allow_subscribe != self.allow_subscribe:
            return False
        if other.disclose_caller != self.disclose_caller:
            return False
        if other.disclose_publisher != self.disclose_publisher:
            return False
        if other.cache != self.cache:
            return False
        if other.created != self.created:
            return False
        if other.owner != self.owner:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def copy(self, other, overwrite=False):
        """
        Copy over other object.

        :param other: Other permission to copy data from.
        :type other: instance of :class:`ManagementRealm`
        :return:
        """
        ConfigurationElement.copy(self, other, overwrite=overwrite)

        if (not self.role_oid and other.role_oid) or overwrite:
            self.role_oid = other.role_oid
        if (not self.uri and other.uri) or overwrite:
            self.uri = other.uri
        if (not self.uri_check_level and other.uri_check_level) or overwrite:
            self.uri_check_level = other.uri_check_level
        if (not self.match and other.match) or overwrite:
            self.match = other.match
        if (not self.allow_call and other.allow_call) or overwrite:
            self.allow_call = other.allow_call
        if (not self.allow_register and other.allow_register) or overwrite:
            self.allow_register = other.allow_register
        if (not self.allow_publish and other.allow_publish) or overwrite:
            self.allow_publish = other.allow_publish
        if (not self.allow_subscribe and other.allow_subscribe) or overwrite:
            self.allow_subscribe = other.allow_subscribe
        if (not self.disclose_caller and other.disclose_caller) or overwrite:
            self.disclose_caller = other.disclose_caller
        if (not self.disclose_publisher and other.disclose_publisher) or overwrite:
            self.disclose_publisher = other.disclose_publisher
        if (not self.cache and other.cache) or overwrite:
            self.cache = other.cache
        if (not self.created and other.created) or overwrite:
            self.created = other.created
        if (not self.owner and other.owner) or overwrite:
            self.owner = other.owner

        # _unknown is not copied!

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ConfigurationElement.marshal(self)

        obj.update(
            {
                "oid": str(self.oid) if self.oid else None,
                "role_oid": str(self.role_oid) if self.role_oid else None,
                "uri": self.uri,
                "uri_check_level": Permission.URI_CHECK_LEVELS_TOSTR[self.uri_check_level]
                if self.uri_check_level
                else None,
                "match": Permission.MATCH_TYPES_TOSTR[self.match] if self.match else None,
                "allow_call": self.allow_call,
                "allow_register": self.allow_register,
                "allow_publish": self.allow_publish,
                "allow_subscribe": self.allow_subscribe,
                "disclose_caller": self.disclose_caller,
                "disclose_publisher": self.disclose_publisher,
                "cache": self.cache,
                "created": int(self.created) if self.created else None,
                "owner": str(self.owner) if self.owner else None,
            }
        )

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in [
                "oid",
                "role_oid",
                "uri",
                "uri_check_level",
                "match",
                "allow_call",
                "allow_register",
                "allow_publish",
                "allow_subscribe",
                "disclose_caller",
                "disclose_publisher",
                "cache",
                "owner",
                "created",
            ]:
                _unknown[k] = data[k]

        role_oid = data.get("role_oid", None)
        assert role_oid is None or type(role_oid) == str
        if role_oid:
            role_oid = UUID(role_oid)

        uri = data.get("uri", None)
        assert uri is None or type(uri) == str

        uri_check_level = data.get("uri_check_level", None)
        assert uri_check_level is None or type(uri_check_level) == str
        if uri_check_level:
            assert uri_check_level in Permission.URI_CHECK_LEVELS_FROMSTR
            uri_check_level = Permission.URI_CHECK_LEVELS_FROMSTR[uri_check_level]

        match = data.get("match", None)
        assert match is None or type(match) == str
        if match:
            assert match in Permission.MATCH_TYPES_FROMSTR
            match = Permission.MATCH_TYPES_FROMSTR[match]

        allow_call = data.get("allow_call", None)
        assert allow_call is None or type(allow_call) == bool

        allow_register = data.get("allow_register", None)
        assert allow_register is None or type(allow_register) == bool

        allow_publish = data.get("allow_publish", None)
        assert allow_publish is None or type(allow_publish) == bool

        allow_subscribe = data.get("allow_subscribe", None)
        assert allow_subscribe is None or type(allow_subscribe) == bool

        disclose_caller = data.get("disclose_caller", None)
        assert disclose_caller is None or type(disclose_caller) == bool

        disclose_publisher = data.get("disclose_publisher", None)
        assert disclose_publisher is None or type(disclose_publisher) == bool

        cache = data.get("cache", None)
        assert cache is None or type(cache) == bool

        created = data.get("created", None)
        assert created is None or type(created) == int
        if created:
            created = np.datetime64(created, "ns")

        owner = data.get("owner", None)
        assert owner is None or type(owner) == str
        if owner:
            owner = UUID(owner)

        obj = Permission(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            role_oid=role_oid,
            uri=uri,
            uri_check_level=uri_check_level,
            match=match,
            allow_call=allow_call,
            allow_register=allow_register,
            allow_publish=allow_publish,
            allow_subscribe=allow_subscribe,
            disclose_caller=disclose_caller,
            disclose_publisher=disclose_publisher,
            cache=cache,
            created=created,
            owner=owner,
            _unknown=_unknown,
        )

        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: user

# CFC organization type
class OrganizationType(object):
    # No organization.
    NONE = 0
    # Personal or individual.
    PERSONAL = 1
    # Company or business.
    BUSINESS = 2
    # Public academics & teaching.
    ACADEMICS = 3



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit

import numpy as np
import pytest
from txaio import (
    time_ns,
    with_twisted,  # noqa
)
from zlmdb import flatbuffers

from cfxdb.xbr import Block


def fill_block(block: Block):
    block.timestamp = np.datetime64(time_ns(), "ns")
    block.block_number = random.randint(0, 2**256 - 1)
    block.block_hash = os.urandom(32)
    block.cnt_events = random.randint(1, 2**32 - 1)


@pytest.fixture(scope="function")
def block():
    _block = Block()
    fill_block(_block)
    return _block


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_block_roundtrip(block, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = block.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 120

    # create python object from bytes (flatbuffes)
    _block = Block.cast(data)

    assert _block.timestamp == block.timestamp
    assert _block.block_number == block.block_number
    assert _block.block_hash == block.block_hash
    assert _block.cnt_events == block.cnt_events


def test_block_roundtrip_perf(block, builder):
    obj = block.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _block = Block.cast(data)
        if True:
            assert _block.timestamp == block.timestamp
            assert _block.block_number == block.block_number
            assert _block.block_hash == block.block_hash
            assert _block.cnt_events == block.cnt_events

            scratch["value"] += int(_block.block_hash[0])
            scratch["value"] += int(_block.block_hash[7])
            scratch["value"] += int(_block.block_hash[19])

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: mrealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Node(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Node()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsNode(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Node
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of this object.
    # Node
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Label for this object (not interpreted by CFC).
    # Node
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Description for this object (not interpreted by CFC).
    # Node
    def Description(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Tags on this object.
    # Node
    def Tags(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Node
    def TagsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Node
    def TagsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Name of this object (must be globally unique within CFC at any given point in time).
    # Also used for WAMP authid under which the node is authenticated on the management realm.
    # Node
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP-cryptosign node public key (32 bytes as HEX encoded string).
    # Node
    def Pubkey(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP ``realm`` the node will be joined on.
    # Node
    def Realm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP-cryptosign ``authid`` the node will be authenticated as.
    # Node
    def Authid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Optional WAMP authextra to be sent to the node when authenticating. CBOR serialized binary.
    # Node
    def Authextra(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Node
    def AuthextraAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Node
    def AuthextraLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Node
    def AuthextraIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0


def NodeStart(builder):
    builder.StartObject(9)


def Start(builder):
    NodeStart(builder)


def NodeAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    NodeAddOid(builder, oid)


def NodeAddLabel(builder, label):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)


def AddLabel(builder, label):
    NodeAddLabel(builder, label)


def NodeAddDescription(builder, description):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0)


def AddDescription(builder, description):
    NodeAddDescription(builder, description)


def NodeAddTags(builder, tags):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(tags), 0)


def AddTags(builder, tags):
    NodeAddTags(builder, tags)


def NodeStartTagsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartTagsVector(builder, numElems: int) -> int:
    return NodeStartTagsVector(builder, numElems)


def NodeAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)


def AddName(builder, name):
    NodeAddName(builder, name)


def NodeAddPubkey(builder, pubkey):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(pubkey), 0)


def AddPubkey(builder, pubkey):
    NodeAddPubkey(builder, pubkey)


def NodeAddRealm(builder, realm):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(realm), 0)


def AddRealm(builder, realm):
    NodeAddRealm(builder, realm)


def NodeAddAuthid(builder, authid):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(authid), 0)


def AddAuthid(builder, authid):
    NodeAddAuthid(builder, authid)


def NodeAddAuthextra(builder, authextra):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(authextra), 0)


def AddAuthextra(builder, authextra):
    NodeAddAuthextra(builder, authextra)


def NodeStartAuthextraVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAuthextraVector(builder, numElems: int) -> int:
    return NodeStartAuthextraVector(builder, numElems)


def NodeEnd(builder):
    return builder.EndObject()


def End(builder):
    return NodeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: meta

# Documentation (text) formats.
class DocFormat(object):
    # Unset documentation format.
    NONE = 0
    # Plain text documentation format (default).
    TEXT = 1
    # Markdown text documentation format.
    MARKDOWN = 2
    # Restructered Text (ReST) text documentation format.
    REST = 3
    # Generic HTML.
    HTML = 4
    # Generic XML.
    XML = 5
    # DocBook XML
    DOCBOOK = 6
    # Jupyter notebook.
    NOTEBOOK = 7



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from zlmdb import MapUuidFlatBuffers, MapUuidUuid, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbrmm import Offer as OfferGen


class _OfferGen(OfferGen.Offer):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsOffer(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _OfferGen()
        x.Init(buf, n + offset)
        return x

    def OfferAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SellerAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def KeyAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ApiAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def UriAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SignatureAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PriceAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Offer(object):
    """
    Data encryption key offerings by XBR providers.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # uint64
        self._timestamp = None

        # [uint8] (uuid)
        self._offer = None

        # [uint8] (address)
        self._seller = None

        # uint64
        self._seller_session_id = None

        # string
        self._seller_authid = None

        # [uint8] (uuid)
        self._key = None

        # [uint8] (uuid)
        self._api = None

        # string
        self._uri = None

        # uint64
        self._valid_from = None

        # [uint8]
        self._signature = None

        # [uint8] (uint256)
        self._price = None

        # [KeyValue]
        self._categories = None

        # uint64
        self._expires = None

        # uint32
        self._copies = None

        # uint32
        self._remaining = None

    def marshal(self) -> dict:
        obj = {
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "offer": self.offer.bytes if self.offer else None,
            "seller": bytes(self.seller) if self.seller else None,
            "seller_session_id": self.seller_session_id,
            "seller_authid": self.seller_authid,
            "key": self.key.bytes if self.key else None,
            "api": self.api.bytes if self.api else None,
            "uri": self.uri,
            "valid_from": int(self.valid_from) if self.valid_from else None,
            "signature": bytes(self.signature) if self.signature else None,
            "price": pack_uint256(self.price) if self.price else 0,
            "categories": self.categories,
            "expires": int(self.expires) if self.expires else None,
            "copies": self.copies,
            "remaining": self.remaining,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def timestamp(self) -> np.datetime64:
        """
        Offer transaction time (epoch time in ns)
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def offer(self) -> uuid.UUID:
        """
        ID of the data encryption key offer.
        """
        if self._offer is None and self._from_fbs:
            if self._from_fbs.OfferLength():
                _offer = self._from_fbs.OfferAsBytes()
                self._offer = uuid.UUID(bytes=bytes(_offer))
        return self._offer

    @offer.setter
    def offer(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._offer = value

    @property
    def seller(self) -> bytes:
        """
        Address of the XBR provider offering the data encryption key.
        """
        if self._seller is None and self._from_fbs:
            if self._from_fbs.SellerLength():
                self._seller = self._from_fbs.SellerAsBytes()
        return self._seller

    @seller.setter
    def seller(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._seller = value

    @property
    def seller_session_id(self) -> int:
        """
        WAMP session ID of the caller that originally placed this offer.
        """
        if self._seller_session_id is None and self._from_fbs:
            self._seller_session_id = self._from_fbs.SellerSessionId()
        return self._seller_session_id

    @seller_session_id.setter
    def seller_session_id(self, value: int):
        assert value is None or type(value) == int
        self._seller_session_id = value

    @property
    def seller_authid(self) -> str:
        """
        WAMP session authid of the caller that originally placed this offer.
        """
        if self._seller_authid is None and self._from_fbs:
            _seller_authid = self._from_fbs.SellerAuthid()
            if _seller_authid:
                self._seller_authid = _seller_authid.decode("utf8")
        return self._seller_authid

    @seller_authid.setter
    def seller_authid(self, value):
        self._seller_authid = value

    @property
    def key(self) -> uuid.UUID:
        """
        ID of the data encryption key offered.
        """
        if self._key is None and self._from_fbs:
            if self._from_fbs.KeyLength():
                _key = self._from_fbs.KeyAsBytes()
                self._key = uuid.UUID(bytes=bytes(_key))
        return self._key

    @key.setter
    def key(self, value):
        assert value is None or isinstance(value, uuid.UUID)
        self._key = value

    @property
    def api(self) -> uuid.UUID:
        """
        ID of the API the encrypted data (this key is for) is provided under.

        :return:
        """
        if self._api is None and self._from_fbs:
            if self._from_fbs.ApiLength():
                _api = self._from_fbs.ApiAsBytes()
                self._api = uuid.UUID(bytes=bytes(_api))
        return self._api

    @api.setter
    def api(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._api = value

    @property
    def uri(self) -> str:
        """
        URI under which the data encrypted with the key offered is provided under.
        """
        if self._uri is None and self._from_fbs:
            _uri = self._from_fbs.Uri()
            if _uri:
                self._uri = _uri.decode("utf8")
        return self._uri

    @uri.setter
    def uri(self, value: str):
        self._uri = value

    @property
    def valid_from(self) -> np.datetime64:
        """
        Timestamp from which the offer is valid (epoch time in ns).
        """
        if self._valid_from is None and self._from_fbs:
            self._valid_from = np.datetime64(self._from_fbs.ValidFrom(), "ns")
        return self._valid_from

    @valid_from.setter
    def valid_from(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._valid_from = value

    @property
    def signature(self) -> bytes:
        """
        Seller delegate signature for the offer. The signature covers all information of the original offer placement request and requestor.
        """
        if self._signature is None and self._from_fbs:
            if self._from_fbs.SignatureLength():
                self._signature = self._from_fbs.SignatureAsBytes()
        return self._signature

    @signature.setter
    def signature(self, value: bytes):
        assert value is None or type(value) == bytes
        self._signature = value

    @property
    def price(self) -> int:
        """
        Price of data encryption key in XBR tokens.
        """
        if self._price is None and self._from_fbs:
            if self._from_fbs.PriceLength():
                _price = self._from_fbs.PriceAsBytes()
                self._price = unpack_uint256(bytes(_price))
            else:
                self._price = 0
        return self._price

    @price.setter
    def price(self, value: int):
        assert value is None or type(value) == int
        self._price = value

    @property
    def categories(self) -> dict:
        """
        Dictionary of optional user defined categories the specific data that is provided falls under.
        """
        if self._categories is None and self._from_fbs:
            num = self._from_fbs.CategoriesKeyLength()
            if num > 0:
                categories = {}
                for i in range(num):
                    key = self._from_fbs.CategoriesKey(i).decode("utf8")
                    value = self._from_fbs.CategoriesValue(i).decode("utf8")
                    categories[key] = value
                self._categories = categories
        return self._categories

    @categories.setter
    def categories(self, values: dict):
        assert values is None or type(values) == dict
        if values:
            assert (type(key) == str for key in values.keys())
            assert (type(value) == str for value in values.values())
        self._categories = values

    @property
    def expires(self) -> np.datetime64:
        """
        Optional data at which this offer expires (epoch time in ns).
        """
        if self._expires is None and self._from_fbs:
            self._expires = np.datetime64(self._from_fbs.Expires(), "ns")
        return self._expires

    @expires.setter
    def expires(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._expires = value

    @property
    def copies(self) -> int:
        """
        Optional maximum number of times this data encryption key is to be sold or 0 for unlimited.
        """
        if self._copies is None and self._from_fbs:
            self._copies = self._from_fbs.Copies()
        return self._copies

    @copies.setter
    def copies(self, value: int):
        assert value is None or type(value) == int
        self._copies = value

    @property
    def remaining(self) -> int:
        """
        Remaining number of copies to be sold (if "copies" is set >0, otherwise 0).
        """
        if self._remaining is None and self._from_fbs:
            self._remaining = self._from_fbs.Remaining()
        return self._remaining

    @remaining.setter
    def remaining(self, value: int):
        assert value is None or type(value) == int
        self._remaining = value

    @staticmethod
    def cast(buf):
        return Offer(_OfferGen.GetRootAsOffer(buf, 0))

    def build(self, builder):
        offer = self.offer.bytes if self.offer else None
        if offer:
            offer = builder.CreateString(offer)

        seller = self.seller
        if seller:
            seller = builder.CreateString(seller)

        seller_authid = self.seller_authid
        if seller_authid:
            seller_authid = builder.CreateString(seller_authid)

        key = self.key.bytes if self.key else None
        if key:
            key = builder.CreateString(key)

        api = self.api.bytes if self.api else None
        if api:
            api = builder.CreateString(api)

        uri = self.uri
        if uri:
            uri = builder.CreateString(uri)

        signature = self.signature
        if signature:
            signature = builder.CreateString(signature)

        price = self.price
        if price:
            price = builder.CreateString(pack_uint256(price))

        categories_keys_vec = None
        categories_values_vec = None

        if self._categories:
            categories_keys = []
            categories_values = []
            for _key, _value in sorted(self._categories.items()):
                assert type(_key) == str, "category key must be string, but was {}: {}".format(type(_key), _key)
                assert type(_value) == str, "category value must be string, but was {}: {}".format(
                    type(_value), _value
                )
                categories_keys.append(builder.CreateString(_key))
                categories_values.append(builder.CreateString(_value))

            OfferGen.OfferStartCategoriesKeyVector(builder, len(categories_keys))
            for _key in categories_keys:
                builder.PrependUOffsetTRelative(_key)
            categories_keys_vec = builder.EndVector()

            OfferGen.OfferStartCategoriesValueVector(builder, len(categories_values))
            for _value in categories_values:
                builder.PrependUOffsetTRelative(_value)
            categories_values_vec = builder.EndVector()

        OfferGen.OfferStart(builder)

        if self.timestamp:
            OfferGen.OfferAddTimestamp(builder, int(self.timestamp))

        if offer:
            OfferGen.OfferAddOffer(builder, offer)

        if seller:
            OfferGen.OfferAddSeller(builder, seller)

        if self.seller_session_id:
            OfferGen.OfferAddSellerSessionId(builder, self.seller_session_id)

        if seller_authid:
            OfferGen.OfferAddSellerAuthid(builder, seller_authid)

        if key:
            OfferGen.OfferAddKey(builder, key)

        if api:
            OfferGen.OfferAddApi(builder, api)

        if uri:
            OfferGen.OfferAddUri(builder, uri)

        if self.valid_from:
            OfferGen.OfferAddValidFrom(builder, int(self.valid_from))

        if signature:
            OfferGen.OfferAddSignature(builder, signature)

        if price:
            OfferGen.OfferAddPrice(builder, price)

        if categories_keys_vec:
            OfferGen.OfferAddCategoriesKey(builder, categories_keys_vec)

        if categories_values_vec:
            OfferGen.OfferAddCategoriesValue(builder, categories_values_vec)

        if self.expires:
            OfferGen.OfferAddExpires(builder, int(self.expires))

        if self.copies:
            OfferGen.OfferAddCopies(builder, self.copies)

        if self.remaining:
            OfferGen.OfferAddRemaining(builder, self.remaining)

        final = OfferGen.OfferEnd(builder)

        return final


@table("dc6d175b-3dd0-4b1f-a6e8-2aec7f0e3fe5", build=Offer.build, cast=Offer.cast)
class Offers(MapUuidFlatBuffers):
    """
    Persisted data encryption key offers.

    Map :class:`zlmdb.MapBytes32FlatBuffers` from ``offer_id`` to :class:`cfxdb.xbr.Offer`
    """


@table("ef5f1cdc-4871-4a03-ac1c-c60e80875b8b")
class IndexOfferByKey(MapUuidUuid):
    """
    Index: key_id -> offer_id
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint

import numpy as np
from zlmdb import MapStringFlatBuffers, flatbuffers, table

from cfxdb.gen.xbrmm import IPFSFile as IPFSFileGen


class _IPFSFileGen(IPFSFileGen.IPFSFile):
    @classmethod
    def GetRootAsIPFSFile(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _IPFSFileGen()
        x.Init(buf, n + offset)
        return x


class IPFSFile:
    """
    Record of downloaded files from Infura
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs
        # string (multihash)
        self._file_hash = None
        # bool
        self._downloaded = None
        # uint32
        self._retries = None
        # uint64 (timestamp)
        self._errored_at = None

    def marshal(self):
        obj = {
            "file_hash": self.file_hash,
            "downloaded": self.downloaded,
            "retries": self.retries,
            "errored_at": self.errored_at,
        }

        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def file_hash(self) -> str:
        """
        IPFS Multihash pointing to a RDF/Turtle file
        """
        if self._file_hash is None and self._from_fbs:
            _hash = self._from_fbs.FileHash()
            if _hash:
                self._file_hash = _hash.decode("utf8")
        return self._file_hash

    @file_hash.setter
    def file_hash(self, value):
        assert value is None or type(value) == str
        self._file_hash = value

    @property
    def downloaded(self) -> bool:
        """
        Whether the file for the given IPFS has was downloaded
        """
        if self._downloaded is None and self._from_fbs:
            self._downloaded = self._from_fbs.Downloaded()
        return self._downloaded

    @downloaded.setter
    def downloaded(self, value: bool):
        assert value is None or type(value) == bool
        self._downloaded = value

    @property
    def retries(self) -> int:
        """
        Number of retries to download the file from Infura
        """
        if self._retries is None and self._from_fbs:
            self._retries = self._from_fbs.Retries()
        return self._retries or 0

    @retries.setter
    def retries(self, value: int):
        assert value is None or type(value) == int
        self._retries = value

    @property
    def errored_at(self) -> np.datetime64:
        """
        Time of last time when downloaded errorred.
        """
        if self._errored_at is None and self._from_fbs:
            self._errored_at = np.datetime64(self._from_fbs.ErroredAt(), "ns")
        return self._errored_at

    @errored_at.setter
    def errored_at(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._errored_at = value

    @staticmethod
    def cast(buf):
        return IPFSFile(_IPFSFileGen.GetRootAsIPFSFile(buf, 0))

    def build(self, builder):
        file_hash = self.file_hash
        if file_hash:
            file_hash = builder.CreateString(file_hash)

        IPFSFileGen.IPFSFileStart(builder)

        if self.errored_at:
            IPFSFileGen.IPFSFileAddErroredAt(builder, int(self.errored_at))

        if self.retries:
            IPFSFileGen.IPFSFileAddRetries(builder, self.retries)

        if file_hash:
            IPFSFileGen.IPFSFileAddFileHash(builder, file_hash)

        IPFSFileGen.IPFSFileAddDownloaded(builder, self.downloaded or False)

        final = IPFSFileGen.IPFSFileEnd(builder)

        return final


@table("def3af4e-190c-4852-93da-731934deef90", build=IPFSFile.build, cast=IPFSFile.cast)
class IPFSFiles(MapStringFlatBuffers):
    """
    Persisted IPFS files download log

    Map :class:`zlmdb.MapStringFlatBuffers` from ``file_hash`` to :class:`cfxdb.xbrmm.IPFSFile`
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class MRealmRun(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MRealmRun()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMRealmRun(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # MRealmRun
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # When the this run ended (Unix time in ns).
    # MRealmRun
    def Ended(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Management realm CFC backend run ID (this is unique over all start-stop cycles of CFC, and constant per run).
    # MRealmRun
    def RunId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MRealmRun
    def RunIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MRealmRun
    def RunIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MRealmRun
    def RunIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # When the this run started (Unix time in ns).
    # MRealmRun
    def Started(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Current CFC management backend state.
    # MRealmRun
    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0


def MRealmRunStart(builder):
    builder.StartObject(4)


def Start(builder):
    MRealmRunStart(builder)


def MRealmRunAddEnded(builder, ended):
    builder.PrependUint64Slot(0, ended, 0)


def AddEnded(builder, ended):
    MRealmRunAddEnded(builder, ended)


def MRealmRunAddRunId(builder, runId):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(runId), 0)


def AddRunId(builder, runId):
    MRealmRunAddRunId(builder, runId)


def MRealmRunStartRunIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartRunIdVector(builder, numElems: int) -> int:
    return MRealmRunStartRunIdVector(builder, numElems)


def MRealmRunAddStarted(builder, started):
    builder.PrependUint64Slot(2, started, 0)


def AddStarted(builder, started):
    MRealmRunAddStarted(builder, started)


def MRealmRunAddState(builder, state):
    builder.PrependUint8Slot(3, state, 0)


def AddState(builder, state):
    MRealmRunAddState(builder, state)


def MRealmRunEnd(builder):
    return builder.EndObject()


def End(builder):
    return MRealmRunEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import numpy as np
import pytest
import txaio
import zlmdb
from zlmdb import flatbuffers

txaio.use_twisted()  # noqa

from txaio import time_ns

from cfxdb.tests._util import _gen_ipfs_hash
from cfxdb.xbr.catalog import Catalog

zlmdb.TABLES_BY_UUID = {}


def fill_catalog(catalog):
    catalog.oid = uuid.uuid4()
    catalog.timestamp = np.datetime64(time_ns(), "ns")
    catalog.seq = random.randint(1, 2**32 - 1)
    catalog.owner = os.urandom(20)
    catalog.terms = _gen_ipfs_hash()
    catalog.meta = _gen_ipfs_hash()
    catalog.tid = os.urandom(32)
    catalog.signature = os.urandom(65)


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


@pytest.fixture(scope="function")
def catalog():
    _catalog = Catalog()
    fill_catalog(_catalog)
    return _catalog


def test_catalog_roundtrip(catalog, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = catalog.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    # assert len(data) == 368

    # create python object from bytes (flatbuffes)
    _catalog = Catalog.cast(data)

    assert _catalog.oid == catalog.oid
    assert _catalog.timestamp == catalog.timestamp
    assert _catalog.seq == catalog.seq
    assert _catalog.owner == catalog.owner
    assert _catalog.terms == catalog.terms
    assert _catalog.meta == catalog.meta
    assert _catalog.tid == catalog.tid
    assert _catalog.signature == catalog.signature


def test_catalog_roundtrip_perf(catalog, builder):
    obj = catalog.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _catalog = Catalog.cast(data)
        if True:
            assert _catalog.oid == catalog.oid
            assert _catalog.timestamp == catalog.timestamp
            assert _catalog.seq == catalog.seq
            assert _catalog.terms == catalog.terms
            assert _catalog.meta == catalog.meta
            assert _catalog.tid == catalog.tid
            assert _catalog.signature == catalog.signature

            scratch["value"] += _catalog.seq

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class MNodeSession(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MNodeSession()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMNodeSession(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # MNodeSession
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # When the his run started (Unix time in ns).
    # MNodeSession
    def Left(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # WAMP session ID of the CF node uplink management session to this CFC instance.
    # MNodeSession
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Management realm CFC backend run ID
    # MNodeSession
    def RunId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MNodeSession
    def RunIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MNodeSession
    def RunIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MNodeSession
    def RunIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # The CF node ID (which is dual-used as WAMP session authid).
    # MNodeSession
    def NodeId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # MNodeSession
    def NodeIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # MNodeSession
    def NodeIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # MNodeSession
    def NodeIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # When the CF node management session joined this CFC instance.
    # MNodeSession
    def Joined(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # When the last heartbeat was received on this CF node session.
    # MNodeSession
    def LastHeartbeat(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0


def MNodeSessionStart(builder):
    builder.StartObject(6)


def Start(builder):
    MNodeSessionStart(builder)


def MNodeSessionAddLeft(builder, left):
    builder.PrependUint64Slot(0, left, 0)


def AddLeft(builder, left):
    MNodeSessionAddLeft(builder, left)


def MNodeSessionAddSession(builder, session):
    builder.PrependUint64Slot(1, session, 0)


def AddSession(builder, session):
    MNodeSessionAddSession(builder, session)


def MNodeSessionAddRunId(builder, runId):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(runId), 0)


def AddRunId(builder, runId):
    MNodeSessionAddRunId(builder, runId)


def MNodeSessionStartRunIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartRunIdVector(builder, numElems: int) -> int:
    return MNodeSessionStartRunIdVector(builder, numElems)


def MNodeSessionAddNodeId(builder, nodeId):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(nodeId), 0)


def AddNodeId(builder, nodeId):
    MNodeSessionAddNodeId(builder, nodeId)


def MNodeSessionStartNodeIdVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartNodeIdVector(builder, numElems: int) -> int:
    return MNodeSessionStartNodeIdVector(builder, numElems)


def MNodeSessionAddJoined(builder, joined):
    builder.PrependUint64Slot(4, joined, 0)


def AddJoined(builder, joined):
    MNodeSessionAddJoined(builder, joined)


def MNodeSessionAddLastHeartbeat(builder, lastHeartbeat):
    builder.PrependUint64Slot(5, lastHeartbeat, 0)


def AddLastHeartbeat(builder, lastHeartbeat):
    MNodeSessionAddLastHeartbeat(builder, lastHeartbeat)


def MNodeSessionEnd(builder):
    return builder.EndObject()


def End(builder):
    return MNodeSessionEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import List, Optional
from uuid import UUID

import numpy as np

from cfxdb.common import ConfigurationElement


class Principal(ConfigurationElement):
    """
    Principals created for use with WAMP authentication. A principal represents the identity
    an application client is authenticated to the application realm joined.

    A principal must have at least one :class:`cfxdb.mrealmschema.Credential` added.
    When an application client connects, it will offer an ``authmethod`` and request a ``realm``
    and (usually) ``authid``. When a matching :class:`cfxdb.mrealmschema.Credential` is found,
    and authentication succeeds using that, the client will be authenticated under the
    :class:`cfxdb.mrealmschema.Principal` associated with the credential.

    .. note::

        It is important to note that while the ``realm`` and ``authid`` requested by the client
        (and defined in the respective :class:`cfxdb.mrealmschema.Credential`) will *usually* be identical
        to the ``realm`` and ``authid`` actually assigned (as defined in the :class:`cfxdb.mrealmschema.Principal`
        associated with the credential), this is allowed to differ in *general*.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        modified: Optional[int] = None,
        arealm_oid: Optional[UUID] = None,
        authid: Optional[str] = None,
        role_oid: Optional[UUID] = None,
        authextra: Optional[dict] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of principal

        :param label: Optional user label of principal

        :param description: Optional user description of principal

        :param tags: Optional list of user tags on principal

        :param modified: Timestamp when the principal was last modified

        :param arealm_oid: ID of the application realm the authenticated principal will be joined to.

        :param authid: WAMP authid of the principal, must be unique within the application realm at any moment in time.

        :param role_oid: ID of the role the authenticated principal will be joined to the application realm.

        :param authextra: Optional authextra information returned to the authenticating principal.
        """
        ConfigurationElement.__init__(self, oid=oid, label=label, description=description, tags=tags)
        self.modified = modified
        self.arealm_oid = arealm_oid
        self.authid = authid
        self.role_oid = role_oid
        self.authextra = authextra

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.modified != self.modified:
            return False
        if other.arealm_oid != self.arealm_oid:
            return False
        if other.authid != self.authid:
            return False
        if other.role_oid != self.role_oid:
            return False
        if other.authextra != self.authextra:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def copy(self, other, overwrite=False):
        """
        Copy over other object.

        :param other: Other principal to copy data from.
        :type other: instance of :class:`cfxdb.mrealm.Principal`
        :return:
        """
        ConfigurationElement.copy(self, other, overwrite=overwrite)

        if (not self.modified and other.modified) or overwrite:
            self.modified = other.modified
        if (not self.arealm_oid and other.arealm_oid) or overwrite:
            self.arealm_oid = other.arealm_oid
        if (not self.authid and other.authid) or overwrite:
            self.authid = other.authid
        if (not self.role_oid and other.role_oid) or overwrite:
            self.role_oid = other.role_oid
        if (not self.authextra and other.authextra) or overwrite:
            self.authextra = other.authextra

        # _unknown is not copied!

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ConfigurationElement.marshal(self)

        obj.update(
            {
                "oid": str(self.oid) if self.oid else None,
                "modified": int(self.modified) if self.modified else None,
                "arealm_oid": str(self.arealm_oid) if self.arealm_oid else None,
                "authid": self.authid,
                "role_oid": str(self.role_oid) if self.role_oid else None,
                "authextra": self.authextra,
            }
        )

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["oid", "modified", "arealm_oid", "authid", "role_oid", "authextra"]:
                _unknown[k] = data[k]

        modified = data.get("modified", None)
        assert modified is None or type(modified) == int
        if modified:
            modified = np.datetime64(modified, "ns")

        arealm_oid = data.get("arealm_oid", None)
        assert arealm_oid is None or type(arealm_oid) == str
        if arealm_oid:
            arealm_oid = UUID(arealm_oid)

        authid = data.get("authid", None)
        assert authid is None or type(authid) == str

        role_oid = data.get("role_oid", None)
        assert role_oid is None or type(role_oid) == str
        if role_oid:
            role_oid = UUID(role_oid)

        authextra = data.get("authextra", None)
        assert authextra is None or type(authextra) == dict

        obj = Principal(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            modified=modified,
            arealm_oid=arealm_oid,
            authid=authid,
            role_oid=role_oid,
            authextra=authextra,
            _unknown=_unknown,
        )

        return obj
